/*********************************************************************
         BLE FlightStick firmware for Adafruit nRF52 feather
                                             V1.40 2021.9.21 by tomo
*********************************************************************/
/* V1.00 2021.07.05 original version
 * V1.10 2021.07.23 modified RE routine 
 * V1.20 2021.08.11 added checking nRF52832 and report Battery charge
 * V1.30 2021.09.04 added Adafruit Feather nRF52840 Express and Inverted Y-axis
 * V1.40 2021.09.21 changed usage class from BLEHidAdafruiut to BLEHidGamepad corresponding to Bluefruit1.0.0
 * V1.50 2021.09.29 setted pnpID with VID_239A(adafruit) & PID_8029(arduino application mode) 
*/
 
#include <bluefruit.h>

//#define NRF52840  // comment out this line if using nRF52832

#ifdef NRF52840
 #include "Adafruit_nRF52840.h" 
#else
 #include "Adafruit_nRF52832.h" //default model
#endif

BLEDis bledis;
BLEHidGamepad blegamepad;
BLEBas blebas;

hid_gamepad_report_t    Gp;     // defined in hid.h from Adafruit_TinyUSB_ArduinoCore
// For Gamepad Button Bit Mask see  hid_gamepad_button_bm_t  typedef defined in hid.h from Adafruit_TinyUSB_ArduinoCore
// For Gamepad Hat    Bit Mask see  hid_gamepad_hat_bm_t     typedef defined in hid.h from Adafruit_TinyUSB_ArduinoCore

// SW Matrix GPIO assign
#define MAXROW 5
const uint8_t PinsROW[MAXROW] = { ROW1,ROW2,ROW3,ROW4,ROW5 };

#define MAXCOL 5
const uint8_t PinsCOL[MAXCOL] = { COL1,COL2,COL3,COL4,COL5 };

// SW Matrix pin asign
#define MAXSW 20 //SW1 - SW20 
const uint8_t RowSW[MAXSW] =
              { ROW3,ROW3,ROW3,ROW3,ROW4,ROW4,ROW4,ROW4,  //Main  SW1 - SW8
                ROW4,ROW3,ROW5,ROW5,ROW5,ROW5,            //AP    SW9 - SW14
                ROW1,ROW2,ROW2,ROW2,ROW2,ROW2 };          //CAM   SW15 - SW20
const uint8_t ColSW[MAXSW] =
              { COL1,COL2,COL3,COL4,COL1,COL2,COL3,COL4,  //Main  SW1 - SW8
                COL5,COL5,COL1,COL2,COL3,COL4,            //AP    SW9 - SW14
                COL5,COL1,COL2,COL3,COL4,COL5 };          //CAM   SW15 - SW20

// Shift SW pin assign
const uint8_t RowShift = ROW5;
const uint8_t ColShift = COL5;

// SW to Button assign
#define SW1    1  //Button1
#define SW2    2  //Button2
#define SW3    3  //Button3
#define SW4    4  //Button4
#define SW5    5  //Button5
#define SW6    7  //Button7
#define SW7    9  //Button9
#define SW8   11  //Button11
#define SW9   13  //Button13
#define SW10  14  //Button14
#define SW11  15  //Button15
#define SW12  17  //Button17
#define SW13  19  //Button19
#define SW14  21  //Button21
#define SW15  24  //Button24
#define SW16  23  //Button23
#define SW17  25  //Button25
#define SW18  27  //Button27
#define SW19  29  //Button29
#define SW20  31  //Button31
const uint8_t SwToButton[MAXSW] =
              { SW1, SW2, SW3, SW4, SW5, SW6, SW7, SW8, SW9, SW10,
                SW11,SW12,SW13,SW14,SW15,SW16,SW17,SW18,SW19,SW20 };

// SW to Button assign at shift SW pushed
#define SW9S  16  //Button16
#define SW10S 18  //Button18
#define SW11S 16  //Button16
#define SW12S 18  //Button18
#define SW13S 20  //Button20
#define SW14S 22  //Button22
#define SW16S 24  //Button24
#define SW17S 26  //Button26
#define SW18S 28  //Button28
#define SW19S 30  //Button30
#define SW20S 32  //Button32
const uint8_t SwToShiftButton[MAXSW] =
              { SW1, SW2, SW3, SW4, SW5, SW6, SW7, SW8, SW9S,SW10S,
                SW11S,SW12S,SW13S,SW14S,SW15,SW16S,SW17S,SW18S,SW19S,SW20S };
const bool    ShiftAvailable[MAXSW]  =
              { false,false,false,false,false,false,false,false,true,true,
                true,true,true,true,false,true,true,true,true,true };

// SW to Reverse Button assign
#define SW5R   6  //Button6
#define SW6R   8  //Button8
#define SW7R  10  //Button10
#define SW8R  12  //Button12
const uint8_t SwToRevButton[MAXSW] =
              { 0, 0, 0, 0,SW5R,SW6R,SW7R,SW8R, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

// HAT SW matrix pin assign                              
#define MAXHAT 4 //HAT1 SW has 4positions (UP,RIGHT,DOWN,LEFT),HAT2 is available when Shift SW is pushed
const uint8_t RowHAT[MAXHAT] = { ROW1,ROW1,ROW1,ROW1 };     //HAT1/2 UP,RIGHT,DOWN,LEFT                  
const uint8_t ColHAT[MAXHAT] = { COL1,COL2,COL3,COL4 };     //HAT1/2 UP,RIGHT,DOWN,LEFT
const uint8_t HATNum[MAXHAT] = { 1, 3, 5, 7 }; 

// Rotary Encoder pin assign
#define MAXRE     2  //2 Rotary encoders are mounted
#define MAX_RE_MILLIS 50     //maximum duration between A and B of Rotary Encorder (msec)
volatile uint32_t REmillis[MAXRE];  //time that Rotary Encorder Pin Falling is detected first
const uint8_t     REpin[MAXRE][2] = { {RE1_A, RE1_B}, {RE2_A, RE2_B} }; //pin Number of Rotary Encoder
volatile bool     REchanged[MAXRE][2] = { false, false, false, false }; //true if Pin Falling is detected
volatile int8_t   REdirec[MAXRE] = { 0, 0 };              //direction of rotation 1:CW, -1:CCW, 0:stop

// Analog ports
// defined in header file

// Adafruit nRF52 model
uint32_t FicrInfoPart;               //0x52832 or 0x52840

// On Board Blue LED port
//#define BLULED   19  //GPIO19

// grobal variables
bool     IsSwOn[MAXSW];       //true while SW is pushed
bool     IsSwLocked[MAXSW];   //true if SW is pushed before Shift SW is pushed
bool     IsShiftOn;           //true while Shift SW pushed
uint32_t TimeShiftChange;     //time(ms) when Shift SW is changed (millis)
#define  SHIFT_LOCK_TIME 1500 //time length(ms) to lock SW after Shift SW is pushed
int8_t   Val_X;               //Xaxis value (-127 to 127)
int8_t   Val_Y;               //Yaxis value (-127 to 127)
int8_t   Val_Z;               //Zaxis value (-127 to 127)
int8_t   Val_Slider1;         //Slider1 value (-127 to 127)
const float Conv = 0.2727;    //ADC compensate coefficient = 256/(1024*3.3/3.6)
#define  HAT_DURATION  300    //duration time(ms) for output after rotation is detected
uint32_t HatMillis[MAXRE] = { 0, 0 };   //time(ms) rotation is detected
uint8_t  Hat3;                //hat3 value (0-8)
uint8_t  Hat4;                //hat4 value (0-8)
const float Vmax = 4.20;      //Battery Voltage of 100% charged
const float Vmin = 3.40;      //Battery Voltage of remaining 0%
#define  REPORT_TURNS  50     //Constant of report counter
uint16_t ReportCount = 0;     //Report counter

void setup() 
{
  Serial.begin(115200);
  delay(50);
  
  FicrInfoPart = *((volatile uint32_t *)FICR_INFO_PART); //read 4 bytes at FICR_INFO_PART addr
  Serial.print("FICR PartCode : ");
  Serial.println(FicrInfoPart,HEX);
  if ( FicrInfoPart != FICR_MODEL ) {
    Serial.println("Warning, the wrong nRF52 model is installed !" );
    delay(1000);
  }
  NfcPubWriteUICR_NFCPINS(); //check NFCPINS
 
  Bluefruit.begin();

  // HID Device can have a min connection interval of 9*1.25 = 11.25 ms
  Bluefruit.Periph.setConnInterval(9, 16); // min = 9*1.25=11.25 ms, max = 16*1.25=20ms
  Bluefruit.setTxPower(4);                 // Check bluefruit.h for supported values
  Bluefruit.setName("nRF52 Flightstick");

  // Configure and Start Device Information Service
  bledis.setManufacturer("Adafruit Industries");
  bledis.setModel("Bluefruit Feather 52");
  char pnpID[] = { 0x01,0x9A,0x23,0x29,0x80,0x01,0x00 }; //SIG_01,VID_239A,PID_8029,REV_0001
  bledis.setPNPID(pnpID,7);
  bledis.begin();

  // Start BLE Battery Service
  blebas.begin();

  // BLE HID
  blegamepad.begin();

  // Set up and start advertising
  // Advertising packet
  Bluefruit.Advertising.addFlags(BLE_GAP_ADV_FLAGS_LE_ONLY_GENERAL_DISC_MODE);
  Bluefruit.Advertising.addTxPower();
  Bluefruit.Advertising.addAppearance(BLE_APPEARANCE_HID_GAMEPAD);

  // Include BLE Gamepad service
  Bluefruit.Advertising.addService(blegamepad);

  // There is enough room for 'Name' in the advertising packet
  Bluefruit.Advertising.addName();

  Bluefruit.Advertising.restartOnDisconnect(true);
  Bluefruit.Advertising.setInterval(32, 244);    // in unit of 0.625 ms
  Bluefruit.Advertising.setFastTimeout(30);      // number of seconds in fast mode
  Bluefruit.Advertising.start(0);                // 0 = Don't stop advertising after n seconds

  Serial.print("Attempting to connect");
  uint8_t i=0;
  while(! Bluefruit.connected()) { 
    Serial.print("."); delay(100);
    if((++i)>50) {
      i=0; //Serial.println();
    }
  };
  delay(1000);//just in case
  Serial.println("\nConnection successful");

  setup_Pinmode();
  reset_RowPins();
  attach_RE_interrupt();
}

void loop() 
{
  set_Buttons();
  set_HAT12();
  set_HAT34();
  set_VR();
  blegamepad.report(&Gp);
  ReportCount += 1;
  if ( ReportCount >= REPORT_TURNS ) {
    ReportCount = 0;
    int16_t  pBatt = read_Battery();
    blebas.write( pBatt );             //report battery charge (%)
    Serial.print( "Batt Cahrge (%) = " );
    Serial.println( pBatt );
  }

  attach_RE_interrupt();
  delay(50);
}

void set_Buttons()
{
  check_ShiftSW();
  read_AllSW();
  
  Gp.buttons = 0;
  for ( uint8_t swNum=0; swNum<MAXSW; swNum++ ) {
    if ( IsSwOn[swNum] && !IsSwLocked[swNum] ) {
      if ( !IsShiftOn ) {
        Gp.buttons += ( 1<<(SwToButton[swNum]-1) );       //turn ON Button if Shift SW is released
      } else {  
        Gp.buttons += ( 1<<(SwToShiftButton[swNum]-1) );  //turn ON shifted-Button if Shift SW is pushed
      }
    }
    if ( !IsSwOn[swNum] && ( SwToRevButton[swNum] > 0 ) ) {
      Gp.buttons += ( 1<<(SwToRevButton[swNum]-1) );      //turn ON reversed-Button if SW is releaseed
    }
  }
}

void check_ShiftSW()
{
  digitalWrite(RowShift, LOW);
  int val = digitalRead( ColShift );
  digitalWrite(RowShift, HIGH);
  if ( !IsShiftOn && val==LOW ) {        //Shift SW turned to ON
    check_SW(true);
    TimeShiftChange = millis();
  } else if ( IsShiftOn && val==HIGH ) { //Shift SW turned to OFF
    check_SW(true);
    TimeShiftChange = millis();
  } else {
    if ( ( millis() - TimeShiftChange ) >= SHIFT_LOCK_TIME ) {
      check_SW(false);
    }
  }
  if ( val ==LOW ) {
    IsShiftOn = true;
  } else {
    IsShiftOn = false;
  }
}

void check_SW( bool lock )
{
  for ( uint8_t swNum=0; swNum<MAXSW; swNum++ ) {
    IsSwLocked[swNum] = false;     //reset SW lock
    if ( lock ) {
      if ( ShiftAvailable[swNum] && IsSwOn[swNum] ) {  //SW was pushed
        IsSwOn[swNum] = false;     //reset SW state
        IsSwLocked[swNum] = true;  //lock SW because SW was pushed
      }
    }
  }
}

void read_AllSW()
{
  reset_RowPins();
  for ( uint8_t swNum=0; swNum<MAXSW; swNum++ ) {
    IsSwOn[swNum] = read_SW( swNum );
  }
}

bool read_SW( uint8_t swNum )
{
  digitalWrite(RowSW[swNum], LOW);
  int val = digitalRead(ColSW[swNum]);
  digitalWrite(RowSW[swNum], HIGH);
  if ( val == LOW ) {
    return true;   //SW on
  } else {
    return false;  //SW off
  }
}

void reset_RowPins()
{
  //set Row Pins High
  for( uint8_t i=0; i<MAXROW; i++ ) {
    digitalWrite( PinsROW[i], HIGH );
  } 
}

void set_HAT12()
{
  uint8_t hatVal = read_HAT();
  digitalWrite(RowShift, LOW);
  int shiftVal = digitalRead( ColShift );
  digitalWrite(RowShift, HIGH);
  if ( shiftVal == HIGH ) {
    Gp.hat1 = hatVal;
    Gp.hat2 = 0;
  } else {
    Gp.hat1 = 0;
    Gp.hat2 = hatVal;
  }
}

uint8_t read_HAT()
{
  for ( uint8_t num=0; num<MAXHAT; num++ ) {
    digitalWrite(RowHAT[num], LOW);
    int val = digitalRead(ColHAT[num]);
    digitalWrite(RowHAT[num], HIGH);
    if ( val == LOW ) {
      return HATNum[num];
    }
  }
  return 0; //HAT off
}

void set_HAT34()
{
  bool shiftON = false;
  digitalWrite(RowShift, LOW);
  int shiftVal = digitalRead( ColShift );
  digitalWrite(RowShift, HIGH);
  if ( shiftVal == LOW ) shiftON = true;
  //HAT3 : RE1
  if ( ( millis() - HatMillis[0] ) > HAT_DURATION ) {
    if ( REdirec[0] > 0 ) {
      if ( !shiftON ) {
        Hat3 = HATNum[0];
      } else {
        Hat3 = HATNum[1];
      }
      HatMillis[0] = millis();
      reset_REparam(0);
    } else if ( REdirec[0] < 0 ) {
      if ( !shiftON ) {
        Hat3 = HATNum[2];
      } else {
        Hat3 = HATNum[3];
      }
      HatMillis[0] = millis();
      reset_REparam(0);
    } else {
      Hat3 = 0;
      HatMillis[0] = 0;
    }    
  }
  Gp.hat3 = Hat3;

  //HAT4 : RE2
  if ( ( millis() - HatMillis[1] ) > HAT_DURATION ) {
    if ( REdirec[1] > 0 ) {
      if ( !shiftON ) {
        Hat4 = HATNum[0];
      } else {
        Hat4 = HATNum[1];
      }
      HatMillis[1] = millis();
      reset_REparam(1);
    } else if ( REdirec[1] < 0 ) {
      if ( !shiftON ) {
        Hat4 = HATNum[2];
      } else {
        Hat4 = HATNum[3];
      }
      HatMillis[1] = millis();
      reset_REparam(1);
    } else {
      Hat4 = 0;
      HatMillis[1] = 0;
    }
  }
  Gp.hat4 = Hat4;
}

void reset_REparam( uint8_t reNum )
{
    REchanged[reNum][0] = false;
    REchanged[reNum][1] = false;
    REdirec[reNum] = 0;
}

void check_RE1_A()
{
  if ( REchanged[0][1] ) {
    if ( (millis()-REmillis[0])<MAX_RE_MILLIS ) {
      REdirec[0] = -1;
    } else {
      REchanged[0][1] = false;
      REchanged[0][0] = true;
      REmillis[0] = millis();
    }
  } else {
    REchanged[0][0] = true;
    REmillis[0] = millis();
  }
}

void check_RE1_B()
{
  if ( REchanged[0][0] ) {
    if ( (millis()-REmillis[0])<MAX_RE_MILLIS ) {
      REdirec[0] = 1;
    } else {
      REchanged[0][0] = false;
      REchanged[0][1] = true;
      REmillis[0] = millis();
    }
  } else {
    REchanged[0][1] = true;
    REmillis[0] = millis();
  }
}

void check_RE2_A()
{
  if ( REchanged[1][1] ) {
    if ( (millis()-REmillis[1])<MAX_RE_MILLIS ) {
      REdirec[1] = -1;
    } else {
      REchanged[1][1] = false;
      REchanged[1][0] = true;
      REmillis[1] = millis();
    }
  } else {
    REchanged[1][0] = true;
    REmillis[1] = millis();
  }
}

void check_RE2_B()
{
  if ( REchanged[1][0] ) {
    if ( (millis()-REmillis[1])<MAX_RE_MILLIS ) {
      REdirec[1] = 1;
    } else {
      REchanged[1][0] = false;
      REchanged[1][1] = true;
      REmillis[1] = millis();
    }
  } else {
    REchanged[1][1] = true;
    REmillis[1] = millis();
  }
}

void attach_RE_interrupt()
{
  attachInterrupt( REpin[0][0], check_RE1_A, FALLING );
  attachInterrupt( REpin[0][1], check_RE1_B, FALLING );
  attachInterrupt( REpin[1][0], check_RE2_A, FALLING );
  attachInterrupt( REpin[1][1], check_RE2_B, FALLING );
}

void detach_RE_interrupt()
{
  detachInterrupt( REpin[0][0] );
  detachInterrupt( REpin[0][1] );
  detachInterrupt( REpin[1][0] );
  detachInterrupt( REpin[1][1] );
}

void set_VR()
{
  //read VR
  Val_X       = read_VR( XAXIS );
  Val_Y       = -read_VR( YAXIS );
  Val_Z       = read_VR( ZAXIS );
  Val_Slider1 = read_VR( SLIDER );
  //set Gp
  Gp.x        = Val_X;
  Gp.y        = Val_Y;
  Gp.z        = Val_Z;
  Gp.slider1  = Val_Slider1;
}

int8_t read_VR( int channel )
{
  float val = analogRead( channel )*Conv - 128;
  if ( val > 127.0 ) {
    return 127;
  } else if ( val < -127.0 ) {
    return -127;
  } else {
    return (int8_t)val;
  }
}

int16_t read_Battery() //return percentage of battery charge
{
  float  vBatt = float( analogRead( BATT ) ) * COEF_BATT;               //CoefBatt is Coefficient to get real Battery Voltage
  int16_t pBatt = (int16_t)( (vBatt - Vmin) * 100.0 / (Vmax - Vmin) ); //convert to percent
  if ( pBatt > 100 ) pBatt = 100;
  if ( pBatt <   0 ) pBatt =   0;
  return pBatt;
}

void setup_Pinmode()
{
  //setup SW matrix
  for ( uint8_t i=0; i<MAXROW; i++ ) {
    pinMode( PinsROW[i], OUTPUT );
  }
  for ( uint8_t i=0; i<MAXCOL; i++ ) {
    pinMode( PinsCOL[i], INPUT_PULLUP );
  }
  for ( uint8_t i=0; i<MAXSW; i++ ) {
    IsSwOn[i] = false;  //reset SW state
  }

  //setup Shift SW
  pinMode( ColShift, INPUT_PULLUP );
  IsShiftOn = false;    //reset Shift SW state

  //setup Rotary Encoder
  for ( uint8_t i=0; i<MAXRE; i++ ) {
    for ( uint8_t j=0; j<2; j++ ) {
      pinMode( REpin[i][j], INPUT_PULLUP );
    }
  }

  //setup ADC
  analogReference( AR_INTERNAL );
  analogReadResolution(10);

  // Reset buttons
  Gp.x       = 0;
  Gp.y       = 0;
  Gp.z       = 0;
  Gp.rz      = 0;     //unused
  Gp.rx      = 0;     //unused
  Gp.ry      = 0;     //unused
  Gp.slider1 = 0x80;
  Gp.slider2 = 0x80;  //unused
  Gp.hat1    = 0;
  Gp.hat2    = 0;
  Gp.hat3    = 0;
  Gp.hat4    = 0;
  Gp.buttons = 0;

  //Blue LED off
  pinMode(BLULED, OUTPUT);
  digitalWrite(BLULED, LOW);
}

/*=================================*/
/*Routine for using NFC pin as GPIO*/
void NfcPubWriteUICR_NFCPINS()
{
    uint32_t nNewVal = 0xFFFFFFFE;
    uint32_t UicrNfcpins = NRF_UICR->NFCPINS; //read 4 bytes at UICR_NFCPINS addr
    Serial.print("current UICR NFCPINS : 0x");
    Serial.println(UicrNfcpins,HEX);
    if ( UicrNfcpins == nNewVal ) return;
    
    /* Turn on flash write enable and wait until the NVMC is ready: */
    NRF_NVMC->CONFIG = (NVMC_CONFIG_WEN_Wen << NVMC_CONFIG_WEN_Pos);
    while (NRF_NVMC->READY == NVMC_READY_READY_Busy) { }

    NRF_UICR->NFCPINS = nNewVal;

    /* Turn off flash write enable and wait until the NVMC is ready: */
    NRF_NVMC->CONFIG = (NVMC_CONFIG_WEN_Ren << NVMC_CONFIG_WEN_Pos);
    while (NRF_NVMC->READY == NVMC_READY_READY_Busy) { }

    UicrNfcpins = NRF_UICR->NFCPINS; //read 4 bytes at UICR_NFCPINS addr
    Serial.print("new UICR NFCPINS : 0x");
    Serial.println(UicrNfcpins,HEX);
    
    return;
}
